/* superlu_dist_config.h.in */

/* Enable CUDA */
/* #undef HAVE_CUDA */

/* Enable NVSHMEM */
/* #undef HAVE_NVSHMEM */

/* Enable HIP */
/* #undef HAVE_HIP */

/* Enable parmetis */
/* #undef HAVE_PARMETIS */

/* Enable colamd */
/* #undef HAVE_COLAMD */

/* Enable LAPACK */
#define SLU_HAVE_LAPACK TRUE

/* Enable CombBLAS */
/* #undef HAVE_COMBBLAS */

/* Enable MAGMA */
/* #undef HAVE_MAGMA */

/* enable 64bit index mode */
/* #undef XSDK_INDEX_SIZE */

#if defined(XSDK_INDEX_SIZE) && (XSDK_INDEX_SIZE == 64)
#define _LONGINT 1
#endif
